/**
 * 
 */
package com.materna.federatedcloudsim.federation;

import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import com.materna.federatedcloudsim.CloudServiceProvider;

/**
 * @author Marc Spohr
 *
 */
public class ServiceProviderRegistry {

	private final static ServiceProviderRegistry instance = new ServiceProviderRegistry();
	private final Set<CloudServiceProvider> cspList;
	private final Set<Service> serviceList;
	private final Map<CloudServiceProvider, Set<Service>> cspServices;
	private final Map<Service, Set<CloudServiceProvider>> serviceCsps;

	/**
	 * Private Constructor, class should be singleton
	 */
	private ServiceProviderRegistry() {
		cspList = new TreeSet<CloudServiceProvider>();
		serviceList = new TreeSet<Service>();
		cspServices = new TreeMap<CloudServiceProvider, Set<Service>>();
		serviceCsps = new TreeMap<Service, Set<CloudServiceProvider>>();
	}

	/**
	 * @return ServiceProviderRegistry global instance
	 */
	public static ServiceProviderRegistry getInstance() {
		return instance;
	}

	/**
	 * @param csp
	 *            the corresponding {@link CloudServiceProvider} for returned
	 *            Service-Set
	 * @return Set of {@link Service} this CSP is offering
	 */
	public Set<Service> getCspServices(CloudServiceProvider csp) {
		return getCspServices().get(csp);
	}

	/**
	 * @param service
	 *            the corresponding {@link Service} for returned CSP-Set
	 * @return Set of {@link CloudServiceProvider} offering this Service
	 */
	public Set<CloudServiceProvider> getServiceCsps(Service service) {
		return getServiceCsps().get(service);
	}

	/**
	 * @param csp
	 *            the {@link CloudServiceProvider} to register
	 * @return {@code true} if CSP has not been registered yet
	 */
	public boolean registerCsp(CloudServiceProvider csp) {
		return (csp != null) && getCspList().add(csp) && (getCspServices().put(csp, new TreeSet<Service>()) == null);
	}

	/**
	 * @param service
	 *            the {@link Service} to register
	 * @return {@code true} if Service has not been registered yet
	 */
	public boolean registerService(Service service) {
		return (service != null) && getServiceList().add(service)
				&& (getServiceCsps().put(service, new TreeSet<CloudServiceProvider>()) == null);
	}

	/**
	 * @param csp
	 *            the {@link CloudServiceProvider} offering a Service
	 * @param service
	 *            the {@link Service} offered by CSP
	 * @return {@code true} if Service and CSP Mapping has been successfully
	 *         added and didn't exist before
	 */
	public boolean registerCspService(CloudServiceProvider csp, Service service) {
		// Register CSP and Service (usually should be done automatically)
		// registerCsp(csp);
		// registerService(s);
		return (csp != null) && (service != null)
				&& (getCspServices(csp).add(service) && getServiceCsps(service).add(csp));
	}

	/**
	 * @param csp
	 *            the {@link CloudServiceProvider} not longer offering a
	 *            specified Service
	 * @param service
	 *            the {@link Service} not longer offered by CSP
	 * @return {@code true} if Service and CSP successfully has been removed
	 *         from mappings
	 */
	public boolean deregisterCspService(CloudServiceProvider csp, Service service) {
		return (csp != null) && (service != null)
				&& (getCspServices(csp).remove(service) && getServiceCsps(service).remove(csp));
	}

	/**
	 * @return the Set of known {@link CloudServiceProvider}
	 */
	public Set<CloudServiceProvider> getCspList() {
		return cspList;
	}

	/**
	 * @return the Set of known {@link Service}
	 */
	public Set<Service> getServiceList() {
		return serviceList;
	}

	/**
	 * @return the {@link CloudServiceProvider} -> provided {@link Service}
	 *         Mapping
	 */
	public Map<CloudServiceProvider, Set<Service>> getCspServices() {
		return cspServices;
	}

	/**
	 * @return the {@link Service} -> provided by {@link CloudServiceProvider}
	 *         Mapping
	 */
	public Map<Service, Set<CloudServiceProvider>> getServiceCsps() {
		return serviceCsps;
	}
}
