/**
 * 
 */
package com.materna.federatedcloudsim.federation.sla;

import java.util.Map;
import java.util.TreeMap;
import java.util.Set;
import java.util.TreeSet;

import org.cloudbus.cloudsim.Log;

import com.materna.federatedcloudsim.CloudServiceProvider;
import com.materna.federatedcloudsim.federation.Service;

/**
 * @author Marc Spohr, Marco Pfahlberg
 * 
 * Represents the Service Level Agreement
 *
 */
public abstract class Sla {
	private static int nextId = 0;
	private final int id;
	private final Map<String, String> slos;
	private CloudServiceProvider providerCsp;
	private Service service;

	private int level; // SLA-Level: 1=gold, 2=silver, 3=bronze
	public static final int LEVEL_GOLD = 1;
	public static final int LEVEL_SILVER = 2;
	public static final int LEVEL_BRONZE = 3;
	private long duration; // how long the sevice takes until finished

	private int migIntraDcCount;
	private int migIntraCspCount;
	private int migIntraFedCount;
	@Deprecated
	private int violationTick;
	private Set<String> brokenSlos;

	private static final long DEFAULT_DURATION = 40320;
	// 120960 minutes ~ 90 days
	// 40320 minutes = 28 days
	// 20160 minutes = 14 days
	// 10080 minutes = 7 days

	/**
	 * Default uptime threshold for gold SLAs
	 */
	private static final float UPTIME_THRESHOLD_GOLD_DEF = (float) 0.999;
	/**
	 * Default uptime threshold for silver SLAs
	 */
	private static final float UPTIME_THRESHOLD_SILVER_DEF = (float) 0.96;
	/**
	 * Default uptime threshold for bronze SLAs
	 */
	private static final float UPTIME_TRHESHOLD_BRONZE_DEF = (float) 0.9;

	/**
	 * Default migration time from host to host within the same datecenter
	 */
	private static final int MIGRATION_TIME_INTRADC_DEF = 5;
	/**
	 * Default migration time from host to host between different datacenters
	 */
	private static final int MIGRATION_TIME_INTERDC_DEF = 10;
	/**
	 * Default migration time from host to host between different cloud service provider
	 */
	private static final int MIGRATION_TIME_INTERCSP_DEF = 20;

	private static float UPTIME_THRESHOLD_GOLD = -1f;
	private static float UPTIME_THRESHOLD_SILVER = -1f;
	private static float UPTIME_TRHESHOLD_BRONZE = -1f;

	private static int MIGRATION_TIME_INTRADC = -1;
	private static int MIGRATION_TIME_INTRA_CSP = -1;
	private static int MIGRATION_TIME_INTRA_FED = -1;

	//private Map<String, Integer> brokenSlos;

	/**
	 * @param providerCsp
	 *            the {@link CloudServiceProvider} offering the service
	 * @param service
	 *            the {@link Service} this SLA specifies
	 *            
	 * calls this(CloudServiceProvider, Service, duration, level) and sets duration and level to default
	 */
	protected Sla(CloudServiceProvider providerCsp, Service service) {
		this(providerCsp, service, Sla.DEFAULT_DURATION, LEVEL_BRONZE);
	}

// TODO Remove unused code found by UCDetector
// 	/**
// 	 * calls this(CloudServiceProvider, Service, duration, level) and sets duration and level to default
// 	 * 
// 	 * @param providerCsp the {@link CloudServiceProvider} offering the service
// 	 * @param service the {@link Service} this SLA specifies
// 	 * @param duration the duration this SLA 'act'
// 	 */
// 	public Sla(CloudServiceProvider providerCsp, Service service, long duration) {
// 		this(providerCsp, service, duration, LEVEL_BRONZE);
// 	}

	/**
	 * calls this(CloudServiceProvider, Service, long, level) and sets level to default
	 * 
	 * @param providerCsp the {@link CloudServiceProvider} offering the service
	 * @param service the {@link Service} this SLA specifies
	 * @param duration the duration this SLA 'act'
	 * @param level the level this SLA
	 */
	private Sla(CloudServiceProvider providerCsp, Service service, long duration, int level) {
		id = nextId++;
		slos = new TreeMap<String, String>();
		setProviderCsp(providerCsp);
		setService(service);
		setDuration(duration);
		setLevel(level);

		migIntraDcCount = 0;
		migIntraCspCount = 0;
		migIntraFedCount = 0;
		violationTick = 0;

		brokenSlos = new TreeSet<String>();
	}

	/**
	 * Returns the CloudServiceProvider of this SLA
	 * 
	 * @return the provider
	 */
	public CloudServiceProvider getProviderCsp() {
		return providerCsp;
	}

	/**
	 * Sets the CloudServiceProvide to 'provider' only if it isn't already set.
	 * 
	 * @param provider
	 *            the provider to set
	 */
	private void setProviderCsp(CloudServiceProvider provider) {
		if (getProviderCsp() != null)
			Log.printLine("SLA " + getId() + ": Provider already set to \"" + getProviderCsp().getName() + "\"!");
		else {
			this.providerCsp = provider;
			getProviderCsp().getSlaManager().addSla(this);
		}
	}

	/**
	 * Returns the service of this SLA.
	 * 
	 * @return the {@link Service} this SLA is concluded to
	 */
	public Service getService() {
		return service;
	}

	/**
	 * Sets the service of this SLA.
	 * 
	 * @param service
	 *            the {@link Service} to set
	 */
	private void setService(Service service) {
		this.service = service;
	}

	/**
	 * Returns the SLOs of this SLA.
	 * 
	 * @return the SLOs
	 */
	public Map<String, String> getSlos() {
		return slos;
	}

	/**
	 * Reuturn the ID of this SLA.
	 * 
	 * @return the id
	 */
	public int getId() {
		return id;
	}

	/**
	 * Sets the duration this sla exist
	 * 
	 * @param duration the duration
	 */
	public void setDuration(long duration) {
		if (duration > 0)
			this.duration = duration;
	}

	/**
	 * Returns the sla level (gold->1, silver->2 or bronze->3)
	 * 
	 * @return the Level
	 */
	public int getLevel() {
		return level;
	}

	/**
	 * Sets the sla level (gold->1, silver->2 or bronze->3)
	 * 
	 * @param level the level
	 */
	public void setLevel(int level) {
		if (level > 0 && level < 4)
			this.level = level;
	}

	/**
	 * Returns the duration this sla exist
	 * 
	 * @return the duration
	 */
	public long getDuration() {
		return duration;
	}

// TODO Remove unused code found by UCDetector
// 	/**
// 	 * DEPRECATED -> obsolete
// 	 * @return
// 	 */
// 	@Deprecated
// 	public static long getDefaultDuration() {
// 		return DEFAULT_DURATION;
// 	}

// TODO Remove unused code found by UCDetector
// 	/**
// 	 * DEPRECATED -> obsolete
// 	 * @return
// 	 */
// 	@Deprecated
// 	public static float getUptimeThresholdGoldDef() {
// 		return UPTIME_THRESHOLD_GOLD_DEF;
// 	}

// TODO Remove unused code found by UCDetector
// 	/**
// 	 * DEPRECATED -> obsolete
// 	 * @return
// 	 */
// 	@Deprecated
// 	public static float getUptimeThresholdSilverDef() {
// 		return UPTIME_THRESHOLD_SILVER_DEF;
// 	}

// TODO Remove unused code found by UCDetector
// 	/**
// 	 * DEPRECATED -> obsolete
// 	 * @return
// 	 */
// 	@Deprecated
// 	public static float getUptimeThresholdBronzeDef() {
// 		return UPTIME_TRHESHOLD_BRONZE_DEF;
// 	}

	/**
	 * Returns the uptime threshold for gold SLAs.
	 * Returns the threshold which has been set, or the default threshold if not set.
	 * 
	 * @return the threshold
	 */
	public static float getUptimeThresholdGold() {
		float ret = UPTIME_THRESHOLD_GOLD;
		if (ret == -1)
			ret = UPTIME_THRESHOLD_GOLD_DEF;

		return ret;
	}

	/**
	 * Returns the uptime threshold for silver SLAs.
	 * Returns the threshold which has been set, or the default threshold if not set.
	 * 
	 * @return the threshold
	 */
	public static float getUptimeThresholdSilver() {
		float ret = UPTIME_THRESHOLD_SILVER;

		if (ret == -1)
			ret = UPTIME_THRESHOLD_SILVER_DEF;

		return ret;
	}

	/**
	 * Returns the uptime threshold for bronze SLAs.
	 * Returns the threshold which has been set, or the default threshold if not set.
	 * 
	 * @return the threshold
	 */
	public static float getUptimeThresholdBronze() {
		float ret = UPTIME_TRHESHOLD_BRONZE;

		if (ret == -1)
			ret = UPTIME_TRHESHOLD_BRONZE_DEF;

		return ret;
	}

	/**
	 * Returns the intra-datacenter migration time.
	 * Returns the migration time which has been set, or the default time if not set.
	 * 
	 * @return the migration time
	 */
	public static int getMigrationTimeIntraDc() {
		int ret = MIGRATION_TIME_INTRADC;

		if (ret == -1)
			ret = MIGRATION_TIME_INTRADC_DEF;

		return ret;
	}

	/**
	 * Returns the inter-datacenter migration time.
	 * Returns the migration time which has been set, or the default time if not set.
	 * 
	 * @return the migration time
	 */
	public static int getMigrationTimeInterDC() {
		int ret = MIGRATION_TIME_INTRA_CSP;

		if (ret == -1)
			ret = MIGRATION_TIME_INTERDC_DEF;

		return ret;
	}

	/**
	 * Returns the inter-cloudServiceProvider migration time.
	 * Returns the migration time which has been set, or the default time if not set.
	 * 
	 * @return the migration time
	 */
	public static int getMigrationTimeInterCSP() {
		int ret = MIGRATION_TIME_INTRA_FED;

		if (ret == -1)
			ret = MIGRATION_TIME_INTERCSP_DEF;

		return ret;
	}

	/**
	 * Sets the sla parameter.
	 * Called after all parameters have been read from properties file
	 * 
	 * @param slaConfig the configuration map
	 */
	public static void configure(Map<String, String> slaConfig) {
		boolean correct = false;

		float utthGold = UPTIME_THRESHOLD_GOLD;
		float utthSilver = UPTIME_THRESHOLD_SILVER;
		float utthBronze = UPTIME_TRHESHOLD_BRONZE;

		int migdc = MIGRATION_TIME_INTRADC;
		int migCSP = MIGRATION_TIME_INTRA_CSP;
		int migFed = MIGRATION_TIME_INTRA_FED;

		try {
			utthGold = Float.parseFloat(slaConfig.get("uptimeThresholdGold"));
			utthSilver = Float.parseFloat(slaConfig.get("uptimeThresholdSilver"));
			utthBronze = Float.parseFloat(slaConfig.get("uptimeThresholdBronze"));

			migdc = Integer.parseInt(slaConfig.get("migrationTimeDC"));
			migCSP = Integer.parseInt(slaConfig.get("migrationTimeCSP"));
			migFed = Integer.parseInt(slaConfig.get("migrationTimeFed"));

			correct = true;
		} catch (Exception e) {
			correct = false;
		}

		if (correct) {
			UPTIME_THRESHOLD_GOLD = utthGold;
			UPTIME_THRESHOLD_SILVER = utthSilver;
			UPTIME_TRHESHOLD_BRONZE = utthBronze;

			MIGRATION_TIME_INTRADC = migdc;
			MIGRATION_TIME_INTRA_CSP = migCSP;
			MIGRATION_TIME_INTRA_FED = migFed;
		}
	}

	/**
	 * Adds one intra-datacenter migration to the intra-datacenter migration count
	 */
	public void addIntraDcMig() {
		migIntraDcCount++;
	}

	/**
	 * Adds one inter-datacenter migration to the inter-datacenter migration count
	 */
	public void addIntraCspMig() {
		migIntraCspCount++;
	}

	/**
	 * Adds one inter-csp migration to the inter-csp migration count
	 */
	public void addIntraFedMig() {
		migIntraFedCount++;
	}

	/**
	 * Returns the intra-datacenter migration count
	 */
	public int getMigIntraDcCount() {
		return migIntraDcCount;
	}

	/**
	 * Returns the inter-datacenter migration count
	 */
	public int getMigInterDcCount() {
		return migIntraCspCount;
	}

	/**
	 * Returns the inter-csp migration count
	 */
	public int getMigInterCspCount() {
		return migIntraFedCount;
	}

	/**
	 * Returns the broken service level objectives of this sla
	 * (NOT USED/NOT PROVIDED YET)
	 * 
	 * @return the broken slos
	 */
	public Set<String> getBrokenSlos() {
		return brokenSlos;
	}

// TODO Remove unused code found by UCDetector
// 	/**
// 	 * Add one broken service level objectives
// 	 * (NOT USED/NOT PROVIDED YET)
// 	 * 
// 	 * @return the broken slo
// 	 */
// 	public void addBrokenSlos(String slo) {
// 		brokenSlos.add(slo);
// 	}

	/**
	 * Returns the max. allowed downtime of this sla. (Minutes)
	 * @return max. allowed downtime in minutes (float)
	 */
	public float getMaxDownTimeMin() {
		float maxDownTimeMin = 0;

		if (level == 1) //Gold
			maxDownTimeMin = (duration - (duration * Sla.getUptimeThresholdGold()));
		else if (level == 2) //Silver
			maxDownTimeMin = (duration - (duration * Sla.getUptimeThresholdSilver()));
		else if (level == 3) //Bronze
			maxDownTimeMin = (duration - (duration * Sla.getUptimeThresholdBronze()));

		return maxDownTimeMin;
	}

	@Deprecated
	public double calcPenality(int estimatetAdditionalDowntime) {

		long downTime = migIntraDcCount * getMigrationTimeIntraDc() + migIntraCspCount * getMigrationTimeInterDC() + migIntraFedCount
				* getMigrationTimeInterCSP() + violationTick * 5;
		if (downTime > getMaxDownTimeMin()) {
			if (level == 1) //Gold
				return 4 + (downTime - getMaxDownTimeMin()) * 0.04;
			else if (level == 2) //Silver
				return 2 + (downTime - getMaxDownTimeMin()) * 0.02;
			else if (level == 3) //Bronze
				return 1 + (downTime - getMaxDownTimeMin()) * 0.01;
			else {
				return 0;
			}

		} else {
			return 0;
		}

	}

// TODO Remove unused code found by UCDetector
// 	@Deprecated
// 	public int isBroken() {
// 		if (slaBroken) {
// 			return 1;
// 		} else
// 			return 0;
// 	}
}
