package com.materna.federatedcloudsim.monitoring;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import org.cloudbus.cloudsim.Host;
import org.cloudbus.cloudsim.Log;

import com.materna.federatedcloudsim.CloudServiceProvider;
import com.materna.federatedcloudsim.FederatedDatacenter;
import com.materna.federatedcloudsim.FederatedHost;
import com.materna.federatedcloudsim.federation.ServiceProviderRegistry;
import com.opencsv.CSVReader;
import com.opencsv.CSVWriter;

//Library for csv management: http://opencsv.sourceforge.net/

public class GephiGenerator {

	// directory to save the gephi-files and where the MigrationLog is stored
	// (set via main.properties)
	private String directory;

	public GephiGenerator(String directory) {
		this.directory = directory;
	}

	/**
	 * generates the GEPHI-Files.
	 * 
	 * Method called in Monitoring (triggered via event at the end of
	 * simulation)
	 */
	protected void generateGephi() throws IOException {

		File migLog = new File(directory + "migration_log.log");

		if (migLog.exists()) {
			generateNodes();
			generateEdges();
		} else {
			Log.printLine("");
			Log.printLine("");
			Log.printLine(
					"[WARNING] Gephi-Files not generated! (File \"migration_log.log\" not found --> maybe migration logging turned off)");
			Log.printLine("");
			Log.printLine("");
		}

	}

	private void generateEdges() throws IOException {
		// variables
		String src = "";
		String dest = "";
		String srcDc = "";
		String destDc = "";
		int weight = 0;
		String pair = "";
		String vm = "";
		int migrations = 0;
		int totalMigrations = 0;
		int percent = 0;

		Hashtable<String, Integer> edges = new Hashtable<String, Integer>();

		TreeMap<String, Integer> vmMigrations = new TreeMap<String, Integer>();

		// read csv file
		String csvFilename = directory + "migration_log.log";
		CSVReader csvReader = new CSVReader(new FileReader(csvFilename), ',', ' ', 1);

		String[] row = null;

		// create new edge file
		String csv = directory + "edges.csv";
		CSVWriter writer = new CSVWriter(new FileWriter(csv));

		// write header
		String[] line = ("Source#Target#Label#Weight").split("#");
		writer.writeNext(line);

		PrintWriter migrationCounter = new PrintWriter(directory + "MigrationCounter.txt", "UTF-8");

		// skip first line (headline)
		csvReader.readNext();

		// go through all lines, find new edges and write them to edges.csv
		while ((row = csvReader.readNext()) != null) {

			src = row[5];
			dest = row[8];
			srcDc = row[4];
			destDc = row[7];
			vm = row[2];

			pair = srcDc + ">" + src + "," + destDc + ">" + dest;

			// if edge already exists: increase weight by 1
			if (edges.containsKey(pair)) {

				weight = edges.get(pair) + 1;

				edges.replace(pair, weight);
			}

			else {

				// add new edge
				edges.put(pair, 1);

			}

			// if vm already exists: increase migration counter by 1
			if (vmMigrations.containsKey(vm)) {

				migrations = vmMigrations.get(vm) + 1;

				vmMigrations.replace(vm, migrations);
			}

			else {

				// add new edge
				vmMigrations.put(vm, 1);

			}

			totalMigrations++;
			src = "";
			dest = "";
			srcDc = "";
			destDc = "";
			weight = 0;
			migrations = 0;

		}

		// enumerate keys and write edges to csv
		Enumeration<String> values = edges.keys();
		while (values.hasMoreElements()) {

			pair = (String) values.nextElement();
			// src = pair.substring(0, 1);
			// dest = pair.substring(2, 3);
			src = pair.substring(0, pair.indexOf(","));
			dest = pair.substring(pair.indexOf(",") + 1, pair.length());
			weight = edges.get(pair);

			// write new edge
			line = (src + "#" + dest + "#" + weight + "#" + weight).split("#");
			writer.writeNext(line);

		}

		// write header for MigrationsCounter
		migrationCounter.println("VM-ID : # migrations : % of total migrations");

		@SuppressWarnings("rawtypes")
		Iterator it = entriesSortedByValues(vmMigrations).iterator();

		while (it.hasNext()) {

			String s = it.next().toString();

			vm = s.substring(0, s.indexOf("="));
			migrations = Integer.parseInt(s.substring(s.indexOf("=") + 1, s.length()));

			percent = Math.round((migrations * 100) / totalMigrations);

			// write new vm entry
			migrationCounter.println(vm + " : " + migrations + " : " + percent + "%");
		}

		totalMigrations = vmMigrations.size();
		migrationCounter.println("");
		migrationCounter.println("---------");
		migrationCounter.println("");
		migrationCounter.println("# of migrated VMs: " + totalMigrations);

		// close file handles
		writer.close();
		csvReader.close();
		migrationCounter.close();
	}

	private static <K, V extends Comparable<? super V>> SortedSet<Map.Entry<K, V>> entriesSortedByValues(
			Map<K, V> map) {

		SortedSet<Map.Entry<K, V>> sortedEntries = new TreeSet<Map.Entry<K, V>>(new Comparator<Map.Entry<K, V>>() {
			@Override
			public int compare(Map.Entry<K, V> e1, Map.Entry<K, V> e2) {
				int res = e1.getValue().compareTo(e2.getValue());
				if (e1.getKey().equals(e2.getKey())) {
					return res; // Code will now handle equality properly
				} else {
					return res != 0 ? res : 1; // While still adding all entries
				}
			}
		});
		sortedEntries.addAll(map.entrySet());
		return sortedEntries;
	}

	private void generateNodes() throws IOException {
		// create new edge file
		String csv = directory + "nodes.csv";
		CSVWriter writer = new CSVWriter(new FileWriter(csv));

		// write header
		String[] line = ("Id#Label").split("#");
		writer.writeNext(line);

		for (CloudServiceProvider csp : ServiceProviderRegistry.getInstance().getCspList())
			for (FederatedDatacenter fedDc : csp.getDcList())
				for (Host host : fedDc.getHostList()) {
					FederatedHost fedHost = ((FederatedHost) host);

					// write new node
					line = (fedHost.getDatacenter().getId() + ">" + fedHost.getId() + "" + "#" + ""
							+ fedHost.getDatacenter().getCsp().getName() + ">" + fedHost.getDatacenter().getId() + ">"
							+ fedHost.getId()).split("#");
					writer.writeNext(line);
				}

		// close file handles
		writer.close();
	}

}
