/**
 * 
 */
package com.materna.federatedcloudsim.monitoring;

import com.materna.federatedcloudsim.core.FederatedCloudSim;


/**
 * @author Marc Spohr
 *
 */
public abstract class ResourcesBean
{
	
	/**
	 * @author Marc Spohr
	 *
	 * @param <T> Data Type contained (Numbers only)
	 */
	public class DataBean<T extends Number>
	{
		private T total;
		private double avg;
		private double lastUpdate;
		private final double startTime;
		
		/**
		 * initialites avg and lastupdate 
		 * @param init initial value
		 */
		public DataBean(T init)
		{
			startTime = FederatedCloudSim.clock();
			setLastUpdate();
			setTotal(init);
			setAvg(init.doubleValue());
		}
		
		/**
		 * recalculates Average
		 */
		protected void updateAvg()
		{
			if(getTotal() != null)
				setAvg(getTotal().doubleValue());
		}

		private void setAvg(double avg)
		{
			double clockMax = FederatedCloudSim.clock() - getStartTime();
			double clockOld = getLastUpdate() - getStartTime();
			double clockNew = FederatedCloudSim.clock() - getLastUpdate();
			this.avg = clockMax > 0 ? (this.avg * clockOld + avg * clockNew) / clockMax : avg;
			setLastUpdate();
		}

		/**
		 * @return average Value
		 */
		public double getAvg()
		{
			updateAvg();
			return this.avg;
		}

		/**
		 * @return last update time
		 */
		public double getLastUpdate()
		{
			return this.lastUpdate;
		}

		private void setLastUpdate()
		{
			this.lastUpdate = FederatedCloudSim.clock();
		}
		
		/**
		 * @return the start time for this DataBean
		 */
		public double getStartTime()
		{
			return this.startTime;
		}

		/**
		 * @return total Value
		 */
		public T getTotal()
		{
			return this.total;
		}
		
		/**
		 * @param total new total to set, triggers average recalculation
		 */
		protected void setTotal(T total)
		{
			updateAvg();
			this.total = total;
		}
	}
	
	/**
	 * @author Marc Spohr
	 *
	 * @param <T> Data Type Contained (Numbers only)
	 */
	public class UsedDataBean<T extends Number> extends DataBean<T>
	{
		private T used;
		
		/**
		 * @param totalInit initial total value
		 */
		public UsedDataBean(T totalInit)
		{
			super(totalInit);
		}
		
		/**
		 * @param totalInit initial total value
		 * @param usedInit initial used value
		 */
		public UsedDataBean(T totalInit, T usedInit)
		{
			this(totalInit);
			setUsed(usedInit);
		}
		
		protected void updateAvg()
		{
			if(getTotal() != null && getUsed() != null)
				super.setAvg(getPercentage());
		}

		/**
		 * @return percentage used
		 */
		public double getPercentage()
		{
			return getTotal() == null || getUsed() == null || getTotal().doubleValue() <= 0 ? 0 : getUsed().doubleValue() / getTotal().doubleValue();
		}
		
		/**
		 * @return used value
		 */
		public T getUsed()
		{
			return this.used;
		}
		
		/**
		 * @param used new used
		 */
		protected void setUsed(T used)
		{
			updateAvg();
			this.used = used;
		}
		
	}
	
	private final DataBean<Integer> peCount;
	private final UsedDataBean<Double> mips;
	private final UsedDataBean<Integer> ram;
	private final UsedDataBean<Long> bw;
	private final UsedDataBean<Long> hostStorage;
	private final DataBean<Integer> vmCount;
	private final DataBean<Integer> cloudletCount;
	private final DataBean<Integer> configuredRam;
	

	/**
	 */
	public ResourcesBean()
	{
		peCount = new DataBean<Integer>(0);
		mips = new UsedDataBean<Double>(0.0, 0.0);
		ram = new UsedDataBean<Integer>(0, 0);
		bw = new UsedDataBean<Long>(0L, 0L);
		hostStorage = new UsedDataBean<Long>(0L, 0L);
		vmCount = new DataBean<Integer>(0);
		cloudletCount = new DataBean<Integer>(0);
		configuredRam = new DataBean<Integer>(0);
	}

	/** Method updating the value of pe
	 */
	protected abstract void updatePeCount();
	
	/** Method updating the value of mips
	 */
	protected abstract void updateMips();
	
	/** Method updating the value of mipsUsed
	 */
	protected abstract void updateMipsUsed();
	
	/** Method updating the value of ram
	 */
	protected abstract void updateRam();
	
	/** Method updating the value of ramUsed
	 */
	protected abstract void updateRamUsed();
	
	/** Method updating the value of bw
	 */
	protected abstract void updateBw();
	
	/** Method updating the value of bwUsed
	 */
	protected abstract void updateBwUsed();
	
	/** Method updating the value of hostStorage
	 */
	protected abstract void updateHostStorage();
	
	/** Method updating the value of hostStorageUsed
	 */
	protected abstract void updateHostStorageUsed();
	
	/** Method updating the value of vmCount
	 */
	protected abstract void updateVmCount();
	
	/** Method updating the value of cloudletCount
	 */
	protected abstract void updateCloudletCount();
	
	/** Method triggers update of parent entities
	 */
	protected abstract void triggerParentUpdate();
	
	protected abstract void updateConfiguredRam();

	
	/**
	 */
	public void update()
	{
		updatePeCount();
		updateMips();
		updateMipsUsed();
		updateRam();
		updateRamUsed();
		updateConfiguredRam();
		updateBw();
		updateBwUsed();
		updateHostStorage();
		updateHostStorageUsed();
		updateVmCount();
		updateCloudletCount();
		triggerParentUpdate();
	}
	
	/**
	 * @return the pe
	 */
	public DataBean<Integer> getPeCount()
	{
		return peCount;
	}
	
	/**
	 * @return the mips
	 */
	public UsedDataBean<Double> getMips()
	{
		return mips;
	}
	
	/**
	 * @return the ram
	 */
	public UsedDataBean<Integer> getRam()
	{
		return ram;
	}
	
	/**
	 * @return the ramConfigured
	 */
	public DataBean<Integer> getConfiguredRam() 
	{
		return configuredRam;
	}

	/**
	 * @return the bw
	 */
	public UsedDataBean<Long> getBw()
	{
		return bw;
	}

	/**
	 * @return the HostStorage
	 */
	public UsedDataBean<Long> getHostStorage()
	{
		return hostStorage;
	}

	/**
	 * @return the vmCount
	 */
	public DataBean<Integer> getVmCount()
	{
		return vmCount;
	}

	/**
	 * @return the cloudletCount
	 */
	public DataBean<Integer> getCloudletCount()
	{
		return cloudletCount;
	}
}
