/**
 * 
 */
package com.materna.federatedcloudsim.scheduling.csp;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import org.cloudbus.cloudsim.DatacenterBroker;
import org.cloudbus.cloudsim.DatacenterCharacteristics;
import org.cloudbus.cloudsim.Log;
import org.cloudbus.cloudsim.UtilizationModel;
import org.cloudbus.cloudsim.UtilizationModelFull;
import org.cloudbus.cloudsim.Vm;
import org.cloudbus.cloudsim.core.CloudSimTags;
import org.cloudbus.cloudsim.core.SimEvent;
import org.cloudbus.cloudsim.lists.VmList;

import com.materna.federatedcloudsim.CloudServiceProvider;
import com.materna.federatedcloudsim.FederatedCloudlet;
import com.materna.federatedcloudsim.FederatedCloudletSchedulerInfinite;
import com.materna.federatedcloudsim.FederatedDatacenter;
import com.materna.federatedcloudsim.FederatedHost;
import com.materna.federatedcloudsim.FederatedVm;
import com.materna.federatedcloudsim.UtilizationModelBitbrains;
import com.materna.federatedcloudsim.UtilizationModelConst;
import com.materna.federatedcloudsim.core.FederatedCloudSim;
import com.materna.federatedcloudsim.core.FederatedCloudSimTags;
import com.materna.federatedcloudsim.core.FederatedConfigurationException;
import com.materna.federatedcloudsim.federation.ServiceRequest;
import com.materna.federatedcloudsim.federation.sla.Sla;
import com.materna.federatedcloudsim.monitoring.Endlog;
import com.materna.federatedcloudsim.monitoring.FileLog;
import com.materna.federatedcloudsim.monitoring.Monitoring;
import com.materna.federatedcloudsim.scheduling.datacenter.FederatedVmAllocationPolicy;
import com.materna.federatedcloudsim.utils.CloudSimConfigurator;
import com.materna.federatedcloudsim.utils.Utils;

/**
 * @author Marc Spohr
 *
 */
public class FederatedDatacenterBroker extends DatacenterBroker {

	private static final String defaultPropFile = "conf" + File.separator + "main.properties";

	private final Properties mainProps;

	private final CloudServiceProvider csp;
	private final Set<ServiceRequest> srAcceptedList;
	private final Set<ServiceRequest> srRejectedList;
	private final Set<Vm> creationRequestedList;
	private FederatedDatacenterBrokerStrategy strategy;

	// private List<FederatedDatacenter> dcList = null;
	private boolean cloudletSubmissionRequested;

	private Set<FederatedVm> rejectedVms;

	/**
	 * @param name
	 *            this DatacenterBroker's name
	 * @param csp
	 *            the {@link CloudServiceProvider} this DatacenterBroker belongs to
	 * @param strategy
	 *            the {@link FederatedDatacenterBrokerStrategy} to use
	 * @throws Exception
	 *             in case of {@code name} contains spaces!
	 */
	public FederatedDatacenterBroker(String name, CloudServiceProvider csp, FederatedDatacenterBrokerStrategy strategy)
			throws Exception {
		super(name);
		this.csp = csp;
		srAcceptedList = new TreeSet<>();
		srRejectedList = new TreeSet<>();
		creationRequestedList = new TreeSet<>();
		setStrategy(strategy);
		// setStrategy(new FederatedDatacenterBrokerStrategyFirstFit());
		// setStrategy(new FederatedDatacenterBrokerStrategyMMDc());
		setCloudletSubmissionRequested(false);
		// mainProps = Utils.loadProperties(defaultPropFile);
		mainProps = CloudSimConfigurator.getInstance().getMainProps();

		rejectedVms = new TreeSet<FederatedVm>();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.cloudbus.cloudsim.DatacenterBroker#processOtherEvent(org.cloudbus .cloudsim.core.SimEvent)
	 */
	@Override
	protected void processOtherEvent(SimEvent ev) {
		switch (ev.getTag()) {
		// Submit cloudlet
		case FederatedCloudSimTags.SERVICE_REQUEST:
			processServiceRequest(ev);
			break;
		case CloudSimTags.CLOUDLET_SUBMIT:
			processCloudletSubmit(ev);
			break;
		case FederatedCloudSimTags.SCHEDULE_REQUEST_DCBROKER:
			processScheduleRequest(ev);
			break;
		case FederatedCloudSimTags.SCHEDULE_REQUEST_CLOUDBROKER_REJ:
			processScheduleRequestRejected(ev);
			break;
		case FederatedCloudSimTags.SCHEDULE_REQUEST_REMOTE:
			processScheduleRequestRemote(ev);
			break;
		case FederatedCloudSimTags.DCBOKER_VM_CHECK:
			processVmCheck(ev);
			break;
		// other unknown tags are processed by this method
		default:
			super.processOtherEvent(ev);
			break;
		}
	}

	/**
	 * 
	 */

	private void processVmCheck(SimEvent ev) {
		FederatedVm[] dataArray = (FederatedVm[]) ev.getData();
		List<FederatedVm> returnedVms = new ArrayList<FederatedVm>();

		for (FederatedVm fedVm : dataArray)
			returnedVms.add(fedVm);

		List<FederatedVm> vms2delete = new ArrayList<FederatedVm>();
		for (FederatedVm fedVm : returnedVms) {
			FederatedDatacenter fedDc = getStrategy().getDc(fedVm);
			if (fedDc != null && fedDc.getVmAllocationPolicy().allocateHostForVm(fedVm)) {
				vms2delete.add(fedVm);
			}
		}

		returnedVms.removeAll(vms2delete);
		rejectedVms.removeAll(vms2delete);

		Map<FederatedVm, FederatedHost> toHost = new TreeMap<>();

		// check if strategy wanna turn on a host...
		if (!returnedVms.isEmpty())
			toHost = getStrategy().checkTurnedOffHosts(returnedVms);

		Set<FederatedHost> hostsStarted = new TreeSet<FederatedHost>();

		for (FederatedVm vm2DelayAlloc : toHost.keySet()) {
			//if host_start not send --> send it
			if (!hostsStarted.contains(toHost.get(vm2DelayAlloc))) {
				FederatedCloudSim.send(this.getId(), toHost.get(vm2DelayAlloc).getDatacenter().getId(), 0,
						FederatedCloudSimTags.HOST_START, toHost.get(vm2DelayAlloc));

				hostsStarted.add(toHost.get(vm2DelayAlloc));
			}

			Map<String, Object> map = new TreeMap<String, Object>();
			map.put("vm", vm2DelayAlloc);
			map.put("host", toHost.get(vm2DelayAlloc));
			
			if(vm2DelayAlloc.getHost() != null)
				vm2DelayAlloc.getHost().getDatacenter().getVmAllocationPolicy().deallocateHostForVm(vm2DelayAlloc);

			// Allocate vm on started host after it started fully
			FederatedCloudSim.send(this.getId(), toHost.get(vm2DelayAlloc).getDatacenter().getId(),
					toHost.get(vm2DelayAlloc).getStatechangeTime() + 1, FederatedCloudSimTags.ALLOCATE_VM_LATER, map);
		}

		returnedVms.removeAll(toHost.keySet());
		rejectedVms.removeAll(toHost.keySet());

		// handle all not allocated vms
		if (!returnedVms.isEmpty()) {
			for(FederatedVm fedV : returnedVms)
				if(fedV.getHost() != null)
					fedV.getHost().getDatacenter().getVmAllocationPolicy().deallocateHostForVm(fedV);
			
			FederatedVm[] vmArray = new FederatedVm[returnedVms.size()];
			send(getCsp().getCb().getId(), 0, FederatedCloudSimTags.SCHEDULE_REQUEST_CLOUDBROKER,
					returnedVms.toArray(vmArray));

			rejectedVms.removeAll(returnedVms);
		}
	}

	public void handleRejectedVmCB(FederatedVm vm) {
		FederatedDatacenter toDc = getStrategy().getDc(vm);
		FederatedHost host = null;

		if (toDc != null) {
			FederatedVmAllocationPolicy destAllocPolicy = (FederatedVmAllocationPolicy) toDc.getVmAllocationPolicy();
			host = destAllocPolicy.findDestinationHost(vm);
			if (host != null)
				destAllocPolicy.allocateHostForVm(vm, host);
		}
		// dc.getResources().update();

		if (toDc == null || host == null)
			FederatedCloudSim.send(getId(), getCsp().getCb().getId(), 0,
					FederatedCloudSimTags.SCHEDULE_REQUEST_CLOUDBROKER, new FederatedVm[] { vm });
	}

	/**
	 * Processes a "SCHEDULE_REQUEST_REMOTE" event. If the vm has been allocated, a "SCHEDULE_REQUEST_REMOTE_ACK" event will be send (to the
	 * ProviderCSP), a "SCHEDULE_REQUEST_REMOTE_REJECT" event (to the ProviderCSP) otherwise.
	 * 
	 * @param ev
	 *            The event data: Contains the fedVm which has to be allocated.
	 */
	private void processScheduleRequestRemote(SimEvent ev) {
		FederatedVm fedVm = (FederatedVm) ev.getData();
		
		FederatedDatacenter possDc = getStrategy().getDc(fedVm);
		FederatedVmAllocationPolicy allocPolicy = null;

		
		
		if (possDc != null) {
			allocPolicy = (FederatedVmAllocationPolicy) possDc.getVmAllocationPolicy();
			if (allocPolicy.allocateHostForVm(fedVm))
				send(ev.getSource(), 0, FederatedCloudSimTags.SCHEDULE_REQUEST_REMOTE_ACK, fedVm);
			else
				send(ev.getSource(), 0, FederatedCloudSimTags.SCHEDULE_REQUEST_REMOTE_REJECT, fedVm);
		} else
			send(ev.getSource(), 0, FederatedCloudSimTags.SCHEDULE_REQUEST_REMOTE_REJECT, fedVm);
	
	fedVm.getFirstCloudlet().getSr().getSla().getProviderCsp().getName();
	}

	@SuppressWarnings("unchecked")
	private void processScheduleRequest(SimEvent ev) {
		ArrayList<FederatedVm> vmList = new ArrayList<FederatedVm>();
		FederatedVm[] retVms = null;

		Object eventData = ev.getData();
		if (eventData instanceof ArrayList<?>)
			vmList = (ArrayList<FederatedVm>) eventData;
		else if (eventData instanceof FederatedVm[])
			retVms = (FederatedVm[]) ev.getData();

		if (retVms != null)
			for (FederatedVm vm : retVms)
				vmList.add(vm);

		FederatedDatacenter dc = (FederatedDatacenter) FederatedCloudSim.getEntity(ev.getSource());
		// FederatedVmAllocationPolicy allocPolicy =
		// (FederatedVmAllocationPolicy) dc.getVmAllocationPolicy();

		List<FederatedVm> allVmsToMove = new ArrayList<FederatedVm>();
		Set<FederatedVm> vmsToMigrate = new TreeSet<FederatedVm>();

		ArrayList<FederatedVm> vmsToMoveCache = getStrategy().findVmsForMigration(vmList, dc);
		vmList.removeAll(vmsToMigrate);
		vmsToMigrate.addAll(vmList); // every Vm which is not part of the method
										// call "findVmsForMigration()", they
										// must be allocated first. (already
										// deallocated)
		vmsToMigrate.addAll(vmsToMoveCache);
		allVmsToMove.addAll(vmsToMigrate);

		for (int i = 0; i < allVmsToMove.size(); i++) {

			FederatedDatacenter toDc = getStrategy().getDc(allVmsToMove.get(i));
			FederatedVmAllocationPolicy srcAllocPolicy = (FederatedVmAllocationPolicy) allVmsToMove.get(i)
					.getFedHostMigHistory().get(allVmsToMove.get(i).getFedHostMigHistory().size() - 1).getDatacenter()
					.getVmAllocationPolicy();

			if (toDc != null) {
				FederatedVmAllocationPolicy destAllocPolicy = (FederatedVmAllocationPolicy) toDc
						.getVmAllocationPolicy();
				FederatedHost host = destAllocPolicy.findDestinationHost(allVmsToMove.get(i));
				if (host != null) {
					srcAllocPolicy.deallocateHostForVm(allVmsToMove.get(i)); // TODO
																				// still
																				// deallocated
					allVmsToMove.get(i).getFedHostMigHistory()
							.get(allVmsToMove.get(i).getFedHostMigHistory().size() - 1).getResources().update();// TODO
																												// still
																												// deallocated
					destAllocPolicy.allocateHostForVm(allVmsToMove.get(i), host);
					allVmsToMove.get(i).getFedHostMigHistory()
							.get(allVmsToMove.get(i).getFedHostMigHistory().size() - 1).getResources().update();

					vmsToMigrate.remove(allVmsToMove.get(i));
				}

			}

		}

		if (vmsToMigrate.size() != 0) {
			
			for(FederatedVm fedV : vmsToMigrate)
				if(fedV.getHost() != null)
					fedV.getHost().getDatacenter().getVmAllocationPolicy().deallocateHostForVm(fedV);
					
			
			FederatedVm[] sendingVms2CB = new FederatedVm[vmsToMigrate.size()];
			sendingVms2CB = vmsToMigrate.toArray(sendingVms2CB);
			FederatedCloudSim.send(getId(), getCsp().getCb().getId(), 0,
					FederatedCloudSimTags.SCHEDULE_REQUEST_CLOUDBROKER, sendingVms2CB);
		}
	}

	/**
	 * This method is invoked after an amount of VMs could not be rescheduled to a federated CSP. They were resend to this DCBroker with a delay.
	 * These VMs are now either scheduled to a DC or they are send back to the CloudBroker to try to find a federated CSP again.
	 * 
	 * @param ev
	 *            the event send from the CloudBroker is the list of rejected VMs
	 */
	private void processScheduleRequestRejected(SimEvent ev) {
		FederatedVm[] retVms = (FederatedVm[]) ev.getData();
		ArrayList<FederatedVm> vmList = new ArrayList<FederatedVm>();
		List<FederatedVm> rejectedVmsCache = new ArrayList<FederatedVm>();

		for (FederatedVm vm : retVms)
			vmList.add(vm);

		for (FederatedVm fedVm : vmList) {

			if (fedVm.isCspRejected()) {
				Map<FederatedVm, FederatedHost> destinations = getStrategy().findPossibleHost(fedVm);

				if (destinations.keySet().size() == 1) {
					if (!destinations.get(fedVm).getDatacenter().getVmAllocationPolicy().allocateHostForVm(fedVm,
							destinations.get(fedVm)))
						rejectedVmsCache.add(fedVm);

					fedVm.setCspReject(false);
				} else if (destinations.keySet().size() > 1) {
					Set<FederatedVm> keys = destinations.keySet();
					keys.remove(fedVm);

					List<FederatedVm> vmsForMigration = new ArrayList<FederatedVm>();

					for (FederatedVm otherVm : keys) {
						otherVm.getFedHostMigHistory().get(otherVm.getFedHostMigHistory().size() - 1).getDatacenter()
								.getVmAllocationPolicy().deallocateHostForVm(otherVm);

						if (destinations.get(otherVm) != null)
							if (!destinations.get(otherVm).getDatacenter().getVmAllocationPolicy()
									.allocateHostForVm(otherVm, destinations.get(otherVm)))
								vmsForMigration.add(otherVm);
							else
								vmsForMigration.add(otherVm);
					}

					FederatedVm[] vmArray = new FederatedVm[vmsForMigration.size()];
					send(getCsp().getCb().getId(), 0, FederatedCloudSimTags.SCHEDULE_REQUEST_CLOUDBROKER,
							vmsForMigration.toArray(vmArray));

					if (!destinations.get(fedVm).getDatacenter().getVmAllocationPolicy().allocateHostForVm(fedVm,
							destinations.get(fedVm)))
						rejectedVmsCache.add(fedVm);

					fedVm.setCspReject(false);
				} else {
					rejectedVmsCache.add(fedVm);

					fedVm.setCspReject(false);
				}
			} else {
				rejectedVmsCache.add(fedVm);
			}

		}

		if (!rejectedVmsCache.isEmpty()) {
			FederatedVm[] vmArray = new FederatedVm[rejectedVmsCache.size()];
			send(getId(), (5 * 60) + 5, FederatedCloudSimTags.DCBOKER_VM_CHECK, rejectedVmsCache.toArray(vmArray));
			rejectedVms.addAll(rejectedVmsCache);
		}
	}

	/**
	 * @param ev
	 *            the {@link SimEvent} to be processed
	 */
	private void processServiceRequest(SimEvent ev) {
		ServiceRequest sr = (ServiceRequest) ev.getData();
		FederatedDatacenter toDc = getStrategy().getDc(sr.getSla(), getDatacenterIdsList());
		
		if (toDc != null) {
			// create cloudlets & vms
			FederatedVm vm = createVms(sr);

			FileLog.getInstance().sr_schedule(sr, null, getCsp(), toDc, vm);
			sendNow(toDc.getId(), CloudSimTags.VM_CREATE_ACK, vm);

			if (sr.getSla().getProviderCsp() == getCsp())
				// Save Request as accepted
				getSrAcceptedList().add(sr);
			else
				// Send Ack to CloudBroker
				sendNow(ev.getSource(), FederatedCloudSimTags.SERVICE_REQUEST_ACC, sr);
			getCreationRequestedList().add(vm);
		} else {
			// Service Request rejected, send to Cloud-Broker
			FileLog.printLine(
					FederatedCloudSim.clock() + ": " + getName() + ": Service " + sr.getSla().getService().getName()
							+ " not schedulable in any local Datacenter" + " ... sending to CloudBroker");
			sendNow(getCsp().getCb().getId(), FederatedCloudSimTags.SERVICE_REQUEST_REJ, sr);
		}
	}

	/**
	 * @param sr
	 *            the {@link ServiceRequest}
	 * @return {@link FederatedVm} created from service request
	 */
	private FederatedVm createVms(ServiceRequest sr) {
		Sla sla = sr.getSla();

		// cloudlet parameters, depending on service & sla
		long length = Long.parseLong(sla.getSlos().get("length"));
		long fileSize = Long.parseLong(sla.getSlos().get("fileSize"));
		long outputSize = Long.parseLong(sla.getSlos().get("outputSize"));
		int pesNumber = Integer.parseInt(sla.getSlos().get("pesNumber"));
		double cpuTime = Double.parseDouble(sla.getSlos().get("cpuUsed"));
		// int memNeeded = Integer.parseInt(sla.getSlos().get("memory"));

		UtilizationModel utilizationModelCpu = new UtilizationModelConst(cpuTime / ((double) length));
		UtilizationModel utilizationModelRam = new UtilizationModelFull();
		try {
			utilizationModelCpu = new UtilizationModelBitbrains(
					mainProps.getProperty("workload.utilization.foldername") + (sr.getId() + 1) + ".csv", 1,
					Integer.parseInt(mainProps.getProperty("workload.utilization.size")));
			utilizationModelRam = new UtilizationModelBitbrains(
					mainProps.getProperty("workload.utilization.foldername") + (sr.getId() + 1) + ".csv", 2,
					Integer.parseInt(mainProps.getProperty("workload.utilization.size")));
		} catch (NumberFormatException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}

		UtilizationModel utilizationModelBw = new UtilizationModelFull();
		FederatedCloudlet c = new FederatedCloudlet(length, pesNumber, fileSize, outputSize, utilizationModelCpu,
				utilizationModelRam, utilizationModelBw, sr);
		c.setUserId(getId());
		c.setEventTime(sr.getRequestTime());
		getCloudletList().add(c);

		// VM Parameters
		long size = 100; // image size (MB)

		// int ram = 512; //vm memory (MB) @author Damian Pasternak
		int ram = Integer.parseInt(sla.getSlos().get("memory")) / pesNumber;

		// workaround: use Mips of hosts for vm mips, since all hosts have same
		// clocking
		int mips = getCsp().getDcList().iterator().next().getHostList().get(0).getPeList().get(0).getMips();
		long bw = 1000;
		String vmm = "Xen"; // VMM name
		FederatedVm vm = new FederatedVm(getId(), mips, pesNumber, ram, bw, size, vmm,
				new FederatedCloudletSchedulerInfinite());
		getVmList().add(vm);
		c.setVm(vm);
		return vm;
	}

	/**
	 * @param ev
	 *            the {@link SimEvent} to be processed
	 */
	private void processCloudletSubmit(SimEvent ev) {
		submitCloudlets();
		setCloudletSubmissionRequested(false);
	}

	/** Overridden Base Class Methods **/

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.cloudbus.cloudsim.DatacenterBroker# processResourceCharacteristicsRequest (org.cloudbus.cloudsim.core.SimEvent)
	 */
	@Override
	protected void processResourceCharacteristicsRequest(SimEvent ev) {
		setDatacenterIdsList(new ArrayList<Integer>());
		setDatacenterCharacteristicsList(new TreeMap<Integer, DatacenterCharacteristics>());

		Log.printLine(FederatedCloudSim.clock() + ": " + getName() + ": Cloud Resource List received with "
				+ getCsp().getDcList().size() + " resource(s)");

		for (FederatedDatacenter dc : getCsp().getDcList()) {
			getDatacenterIdsList().add(dc.getId());
			sendNow(dc.getId(), CloudSimTags.RESOURCE_CHARACTERISTICS, getId());
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.cloudbus.cloudsim.DatacenterBroker#processCloudletReturn(org.cloudbus .cloudsim.core.SimEvent)
	 */
	@Override
	protected void processCloudletReturn(SimEvent ev) {
		FederatedCloudlet cloudlet = (FederatedCloudlet) ev.getData();
		getCloudletReceivedList().add(cloudlet);
		cloudlet.getSr().isCloudletFinished(cloudlet);
		Log.printLine(
				FederatedCloudSim.clock() + ": " + getName() + ": Cloudlet " + cloudlet.getCloudletId() + " received");
		cloudletsSubmitted--;
		sendNow(getVmsToDatacentersMap().get(cloudlet.getVmId()), CloudSimTags.VM_DESTROY,
				VmList.getById(vmList, cloudlet.getVmId()));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.cloudbus.cloudsim.DatacenterBroker#processVmCreate(org.cloudbus. cloudsim .core.SimEvent)
	 */
	@Override
	protected void processVmCreate(SimEvent ev) {
		int[] data = (int[]) ev.getData();
		int datacenterId = data[0];
		int vmId = data[1];
		int result = data[2];

		if (result == CloudSimTags.TRUE) {
			getVmsToDatacentersMap().put(vmId, datacenterId);
			getVmsCreatedList().add(VmList.getById(getVmList(), vmId));
			// Log.printLine(
			// FederatedCloudSim.clock() + ": " + getName() + ": VM #" + vmId +
			// " has been created in Datacenter #"
			// + datacenterId + ", Host #" + VmList.getById(getVmsCreatedList(),
			// vmId).getHost().getId());
			FederatedHost host = (FederatedHost) VmList.getById(getVmsCreatedList(), vmId).getHost();
			host.getResources().update();
			// Log.printLine(FederatedCloudSim.clock() + " VM Count on Host: " +
			// host.getVmList().size());
			// host.getResources().update();

			// Send Cloudlets to New Created VMs
			if (!isCloudletSubmissionRequested()) {
				sendNow(getId(), CloudSimTags.CLOUDLET_SUBMIT);
				setCloudletSubmissionRequested(true);
			}
		} else {
			Log.printLine(FederatedCloudSim.clock() + ": " + getName() + ": Creation of VM #" + vmId
					+ " failed in Datacenter #" + datacenterId);

			// Request Creation in Next DC, if there are more than one
			// if(getDatacenterIdsList().size() > 1)
			// for(int i = 0; i < getDatacenterIdsList().size(); i++)
			// {
			// if(getDatacenterIdsList().get(i) == datacenterId)
			// {
			// i = (i + 1) % getDatacenterIdsList().size();
			// Log.printLine(CloudSim.clock() + ": " + getName() +
			// ": Trying to Create VM #" + vmId
			// + " in DataCenter #" + getDatacenterIdsList().get(i));
			// sendNow(getDatacenterIdsList().get(i),
			// CloudSimTags.VM_CREATE_ACK, VmList.getById(vmList, vmId));
			// break;
			// }
			// }

			// no further scheduling, drop!
			// TODO: casting!! control about more than once cloudlet per vm!!
			FederatedVm vm = (FederatedVm) VmList.getById(getVmList(), vmId);
			// getSrRejectedList().add();
			getCloudletList().remove(vm.getCloudlets().iterator().next());
			getSrAcceptedList().remove(vm.getCloudlets().iterator().next().getSr());
			sendNow(getCsp().getCb().getId(), FederatedCloudSimTags.SERVICE_REQUEST_REJ,
					vm.getCloudlets().iterator().next().getSr());
		}
	}

	/**
	 * @return the {@link CloudServiceProvider}
	 */
	public CloudServiceProvider getCsp() {
		return csp;
	}

	/**
	 * @return the {@link FederatedDatacenterBrokerStrategy}
	 */
	public FederatedDatacenterBrokerStrategy getStrategy() {
		return strategy;
	}

	/**
	 * @param strategy
	 *            the {@link FederatedDatacenterBrokerStrategy} to use ( {@code null} to use {@link FederatedDatacenterBrokerStrategySimple})
	 */
	public void setStrategy(FederatedDatacenterBrokerStrategy strategy) {
		if (strategy == null) {
			strategy = new FederatedDatacenterBrokerStrategySimple();
		}
		this.strategy = strategy;
		getStrategy().setDcBroker(this);
	}

	/**
	 * @return the Set of accepted {@link ServiceRequest}
	 */
	public Set<ServiceRequest> getSrAcceptedList() {
		return srAcceptedList;
	}

	/**
	 * @return the Set of rejected {@link ServiceRequest}
	 */
	public Set<ServiceRequest> getSrRejectedList() {
		return srRejectedList;
	}

	/**
	 * @return the Set of {@link Vm} requested for creation
	 */
	public Set<Vm> getCreationRequestedList() {
		return creationRequestedList;
	}

	/**
	 * @return the cloudletSubmissionRequested
	 */
	public boolean isCloudletSubmissionRequested() {
		return cloudletSubmissionRequested;
	}

	/**
	 * @param cloudletSubmissionRequested
	 *            the cloudletSubmissionRequested to set
	 */
	private void setCloudletSubmissionRequested(boolean cloudletSubmissionRequested) {
		this.cloudletSubmissionRequested = cloudletSubmissionRequested;
	}

	public void configure(Map<String, String> params) throws FederatedConfigurationException {

	}
}
