package com.materna.federatedcloudsim.scheduling.federation;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import org.cloudbus.cloudsim.core.SimEvent;

import com.materna.federatedcloudsim.CloudServiceProvider;
import com.materna.federatedcloudsim.FederatedDatacenter;
import com.materna.federatedcloudsim.FederatedHost;
import com.materna.federatedcloudsim.FederatedVm;
import com.materna.federatedcloudsim.core.FederatedCloudSim;
import com.materna.federatedcloudsim.core.FederatedCloudSimTags;
import com.materna.federatedcloudsim.federation.ServiceProviderRegistry;
import com.materna.federatedcloudsim.federation.ServiceRequest;
import com.materna.federatedcloudsim.federation.sla.SlaCsp;
import com.materna.federatedcloudsim.scheduling.datacenter.FederatedVmAllocationPolicy;

/**
 * This class is a CloudBrokerStrategy the federation CloudServiceProviders
 * costlevel. When a VM is to be migrated, the Strategy initiates an auction. As
 * destination for the strategic migration the auction winner with the cheapest
 * offer is selected.
 * 
 * The Strategy accepts migration requests and auction invitations, if a
 * suitable host is found by evoking the DataCenterBrokers getDc(fedVm) and the
 * related VmAllocationPolicies findDestinationHost(fedVm) method.
 * 
 * @author Marcel Krger
 */
public class AuctionStrategy2RoundVmCheapestFit extends AuctionStrategy {

	private final int maxBiddingRounds = 2;

	public AuctionStrategy2RoundVmCheapestFit(String auctionForm) {
		super(auctionForm);
		getBiddingParameters().put("cpuCostFactor", "");
		getBiddingParameters().put("ramCostFactor", "");
		getBiddingParameters().put("resourceSlaFactorSilver", "");
		getBiddingParameters().put("resourceSlaFactorBronze", "");
	}

	@Override
	public void startInvitation(Auction newAuction) {

		for (CloudServiceProvider csp : ServiceProviderRegistry.getInstance().getCspList()) {

			if (csp != getCb().getCsp() && csp.getCb().getStrategy().isAuctionBased()) {
				FederatedCloudSim.send(getCb().getId(), csp.getCb().getId(), 0d,
						FederatedCloudSimTags.AUCTION_INVITATION, newAuction);
				if (getAuctionManager().getAuctionInvitationMap().containsKey(newAuction)) {
					getAuctionManager().getAuctionInvitationMap().get(newAuction).add(csp);
				} else {
					List<CloudServiceProvider> list = new ArrayList<>();
					list.add(csp);
					getAuctionManager().getAuctionInvitationMap().put(newAuction, list);
				}
			}
		}
		FederatedCloudSim.send(getCb().getId(), getCb().getId(), 0, FederatedCloudSimTags.CHECK_INVITATION, newAuction);
	}

	// In this method a strategy might react to an invitation according to the
	// DCs Utilization. So far it checks whether it is space at all in a
	// possible DC
	@Override
	public boolean handleInvitation(Auction auction) {
		String auctionForm = auction.getAuctionForm();
		if (auctionForm.equals(getAuctionForm())) {
			FederatedHost fedHost = null;
			FederatedVm fedVm = (FederatedVm) auction.getAuctionObject();
			FederatedDatacenter possDc = this.getCb().getCsp().getDcBroker().getStrategy().getDc(fedVm);

			if (possDc != null) {
				FederatedVmAllocationPolicy possAllocPolicy = (FederatedVmAllocationPolicy) possDc
						.getVmAllocationPolicy();
				fedHost = possAllocPolicy.findDestinationHost(fedVm);
			}
			boolean retVal = fedHost != null ? true : false;
			if (!retVal){
				getAuctionManager().increaseRejectedInvitations();
			}
			return retVal;
		}
		getAuctionManager().increaseRejectedInvitations();
		return false;
	}

	@Override
	public void startAuction(Auction auction) {

		if (auction.getParticipantList().isEmpty()) {
			// No provider available for auction, send rejected vm to DCBroker
			FederatedVm vm = (FederatedVm) auction.getAuctionObject();
			ArrayList<FederatedVm> vmToMove = new ArrayList<FederatedVm>();
			vmToMove.add(vm);
			getAuctionManager().getRejectedAuctionsList().add(auction);
			getAuctionManager().increaseRejectedLevelOne();
			FederatedCloudSim.send(getCb().getId(), auction.getAuctioneer().getDcBroker().getId(), 0,
					FederatedCloudSimTags.SCHEDULE_REQUEST_CLOUDBROKER_REJ, vmToMove.toArray(new FederatedVm[1]));
		} else {
			for (CloudServiceProvider csp : auction.getParticipantList()) {
				AuctionMessage message = new AuctionMessage(auction, csp);
				message.setBiddingParameters(getBiddingParameters());
				FederatedCloudSim.send(getCb().getId(), csp.getCb().getId(), 0, FederatedCloudSimTags.AUCTION_MESSAGE,
						message);
				auction.getMessageMap().put(csp, message);
			}
			FederatedCloudSim.send(getCb().getId(), getCb().getId(), 0,
					FederatedCloudSimTags.CHECK_AUCTION_MESSAGE_RETURN, auction);
		}
	}

	@Override
	public void handleAuctionMessage(AuctionMessage message) {

		getBiddingStrategy().calcBiddingParameter(message);

		// Reserviere platz fr die potentiell zu erhaltende Vm, oder nicht
		// (Strategische Entscheidung)

	}

	@Override
	public void finishAuctionRound(Auction auction) {
		FederatedVm vm = (FederatedVm) auction.getAuctionObject();
		if (auction.getParticipantList().isEmpty()) {
			auction.setFinished(true);
			getAuctionManager().getRejectedAuctionsList().add(auction);
			getAuctionManager().increaseRejectedLevelTwo();
			// Auction is rejected from all participants -> reschedule Vm
			FederatedVm[] vmArray = { vm };
			FederatedCloudSim.send(getCb().getId(), getCb().getCsp().getDcBroker().getId(), 0,
					FederatedCloudSimTags.SCHEDULE_REQUEST_CLOUDBROKER_REJ, vmArray);
			getCb().getVmsForMigration().remove(vm);

		} else {
			AuctionMessage oneMessage = auction.getMessageMap().get(auction.getParticipantList().get(0));
			int round = auction.setRound(oneMessage.getRound());
			Map<CloudServiceProvider, Double> bidValueMap = new TreeMap<>();
			bidValueMap.putAll(valueBiddingParameter(auction));
			List<CloudServiceProvider> sortedList = new ArrayList<>();
			sortedList.addAll(sortMapAscendingByBidValue(bidValueMap));
			auction.getBestBidByRoundMap().put(round, bidValueMap.get(sortedList.get(0)));
			round++;
			if (round <= maxBiddingRounds) {
				auction.getMessageMap().get(sortedList.get(0)).setLeadingOffer(true);
				auction.getBidList().clear();
				for (CloudServiceProvider participant : auction.getParticipantList()) {
					AuctionMessage message = auction.getMessageMap().get(participant);
					message.setRound(round);
					FederatedCloudSim.send(getCb().getId(), participant.getCb().getId(), 0,
							FederatedCloudSimTags.AUCTION_MESSAGE, message);
				}
				FederatedCloudSim.send(getCb().getId(), getCb().getId(), 0,
						FederatedCloudSimTags.CHECK_AUCTION_MESSAGE_RETURN, auction);
			} else {
				CloudServiceProvider winner = sortedList.get(0);
				auction.setFinished(true);
				auction.setWinner(winner);
				getAuctionManager().getCompletedAuctionsList().add(auction);
				for (CloudServiceProvider csp : auction.getParticipantList())
					auction.getMessageMap().get(csp).setLeadingOffer(false);
				auction.getMessageMap().get(winner).setLeadingOffer(true);
				vm.setAuctioned(true);
				FederatedCloudSim.send(getCb().getId(), winner.getCb().getId(), 0,
						FederatedCloudSimTags.MIGRATION_REMOTE, vm);
			}
		}
	}

	public List<CloudServiceProvider> sortMapAscendingByBidValue(Map<CloudServiceProvider, Double> bids) {
		
		Map<CloudServiceProvider, Double> cacheMap = new TreeMap<CloudServiceProvider, Double>(bids);
		List<CloudServiceProvider> resultList = new ArrayList<>();
		CloudServiceProvider cheapest = null;
		Double leastBid = Double.MAX_VALUE;
		while (!cacheMap.isEmpty()) {
			for (Map.Entry<CloudServiceProvider, Double> entry : cacheMap.entrySet()) {
				if (entry.getValue() < leastBid) {
					leastBid = entry.getValue();
					cheapest = entry.getKey();
				}
			}
			resultList.add(cheapest);
			cacheMap.remove(cheapest);
			cheapest = null;
			leastBid = Double.MAX_VALUE;
		}
		return resultList;
	}

	public Map<CloudServiceProvider, Double> valueBiddingParameter(Auction auction) {
		Map<CloudServiceProvider, Double> bidValueMap = new TreeMap<>();
		for (CloudServiceProvider csp : auction.getBidList().keySet()) {
			FederatedVm vm = (FederatedVm) auction.getAuctionObject();
			// Sla level Siver = 2, Sla level Bronze = 3, Gold Sla is not
			// supposed to be migrated
			int slaLevel = vm.getFirstCloudlet().getSr().getSla().getLevel();
			String slaFactorString = "";
			switch (slaLevel) {
			case 2:
				slaFactorString = "resourceSlaFactorSilver";
			case 3:
				slaFactorString = "resourceSlaFactorBronze";
			default:
				slaFactorString = "resourceSlaFactorSilver";
			}
			Double cpuFactor = Double.parseDouble(auction.getBidList().get(csp).get("cpuCostFactor"));
			Double ramFactor = Double.parseDouble(auction.getBidList().get(csp).get("ramCostFactor"));
			Double slaFactor = Double.parseDouble(auction.getBidList().get(csp).get(slaFactorString));
			Double value = cpuFactor * slaFactor + ramFactor * slaFactor;
			bidValueMap.put(csp, value);
		}
		return bidValueMap;
	}

	@Override
	public CloudServiceProvider getCsp(ServiceRequest sr, int evSource) {
		// first CSP with existing SLA is taken
		Set<SlaCsp> cspSlas = getCb().getCsp().getSlaManager().getUtilizedServiceSlas(sr.getSla().getService());
		Iterator<SlaCsp> iterator = cspSlas != null ? cspSlas.iterator() : null;
		CloudServiceProvider csp;
		while (iterator != null && iterator.hasNext()) {
			csp = iterator.next().getProviderCsp();
			// Event sent by own DC-Broker -> schedule to next available CSP
			if ((evSource == getCb().getCsp().getDcBroker().getId())) {
				return csp;
			}
			// Event sent by other CSPs CB -> execution rejected, schedule to
			// next CSP
			else if ((csp.getCb().getId() == evSource) && iterator.hasNext()) {
				csp = iterator.next().getProviderCsp();
				return csp;
			}
		}
		return null;
	}

	@Override
	public boolean acceptRemoteServiceRequest(ServiceRequest sr) {
		// TODO Auto-generated method stub
		return false;
	}

	@Override
	public double getRescheduleDelay(ServiceRequest sr) {
		// TODO Auto-generated method stub
		return 0;
	}

	/**
	 * @return the maxBiddingRounds
	 */
	public int getMaxBiddingRounds() {
		return maxBiddingRounds;
	}

	@Override
	public CloudServiceProvider findCsp(FederatedVm vm) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public boolean handleMigrationRequest(FederatedVm fedVm) {
		FederatedHost fedHost = null;

		FederatedDatacenter possDc = this.getCb().getCsp().getDcBroker().getStrategy().getDc(fedVm);

		if (possDc != null) {
			FederatedVmAllocationPolicy possAllocPolicy = (FederatedVmAllocationPolicy) possDc.getVmAllocationPolicy();
			fedHost = possAllocPolicy.findDestinationHost(fedVm);
		}

		return fedHost != null ? true : false;
	}

	@Override
	public CloudServiceProvider handleRejectedMigrationRequest(FederatedVm vm) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public void handleFedMigrationAnswer(FederatedVm fedVm, SimEvent ev) {
		// TODO Auto-generated method stub
	}
}
