/**
 * 
 */
package com.materna.federatedcloudsim.scheduling.federation;

import java.util.ArrayList;
import java.util.Map;

import org.cloudbus.cloudsim.core.SimEvent;

import com.materna.federatedcloudsim.CloudServiceProvider;
import com.materna.federatedcloudsim.FederatedVm;
import com.materna.federatedcloudsim.core.FederatedCloudSimTags;
import com.materna.federatedcloudsim.core.FederatedConfigurationException;
import com.materna.federatedcloudsim.federation.Service;
import com.materna.federatedcloudsim.federation.ServiceRequest;

/**
 * @author Marc Spohr
 *
 */
public abstract class CloudBrokerStrategy {
	private CloudBroker cb;
	private boolean isAuctionBased;

	public CloudBrokerStrategy(boolean isAuctionBased) {
		this.isAuctionBased = isAuctionBased;
	}

	/**
	 * @param sr
	 *            the {@link ServiceRequest}
	 * @param evSource
	 *            the id of the processed event source
	 * @return {@link CloudServiceProvider} this ServiceRequest should be
	 *         scheduled to
	 */
	public abstract CloudServiceProvider getCsp(ServiceRequest sr, int evSource);

	public void configure(Map<String, String> params) throws FederatedConfigurationException {
	}

	/**
	 * This method finds, according to the Strategy that is implemented, a
	 * federated {@link CloudServiceProvider} to which a migration request for
	 * the referenced {@link FederatedVm} is send.
	 * 
	 * @param vm
	 *            the VM to be rescheduled on a different CSP
	 * @return {@link CloudServiceProvider} this {@link FederatedVm} should be
	 *         scheduled to
	 */
	public abstract CloudServiceProvider findCsp(FederatedVm vm);

	/**
	 * This method implements the strategic decision whether a migration request
	 * is to accept or decline
	 * 
	 * @param fedVm
	 *            for which the request is send
	 * @return boolean whether to accept
	 */
	public abstract boolean handleMigrationRequest(FederatedVm fedVm);

	/**
	 * !!!optional if strategy is auction based
	 * 
	 * This method implements the logic to choose the next possible Csp, if the
	 * last one refused to accept the request.
	 * 
	 * @param fedVm
	 *            for which the request is rejected
	 * @return {@link CloudServiceProvider} this {@link FederatedVm} should be
	 *         scheduled to next
	 */
	public abstract CloudServiceProvider handleRejectedMigrationRequest(FederatedVm vm);

	/**
	 * optional This method is used to set or reset all used attributes after a
	 * vm is migrated it is used for both for successful and unsuccessful
	 * migrations
	 * 
	 * @param fedVm
	 */
	public abstract void handleFedMigrationAnswer(FederatedVm fedVm, SimEvent ev);

	/**
	 * @param sr
	 *            the {@link ServiceRequest}
	 * @return {@code true} if Service Request should be accepted
	 */
	public abstract boolean acceptRemoteServiceRequest(ServiceRequest sr);

	/**
	 * @param sr
	 *            the {@link ServiceRequest}
	 * @return the delay (in s) when this own ServiceRequest will be rescheduled
	 *         to DatacenterBroker, {@code negative} value means to completely
	 *         reject this ServiceRequest
	 */
	public abstract double getRescheduleDelay(ServiceRequest sr);

	/**
	 * @return the {@link CloudBroker}
	 */
	public CloudBroker getCb() {
		return cb;
	}

	/**
	 * @param cb
	 *            the {@link CloudBroker} to set
	 */
	public void setCb(CloudBroker cb) {
		this.cb = cb;
	}

	/**
	 * @return the isAuctionBased
	 */
	public boolean isAuctionBased() {
		return isAuctionBased;
	}

	/**
	 * @param isAuctionBased
	 *            the isAuctionBased to set
	 */
	public void setAuctionBased(boolean isAuctionBased) {
		this.isAuctionBased = isAuctionBased;
	}

	public void initializeParameter() {

	}
}
