package com.materna.federatedcloudsim.scheduling.federation;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import org.cloudbus.cloudsim.core.SimEvent;

import com.materna.federatedcloudsim.CloudServiceProvider;
import com.materna.federatedcloudsim.FederatedDatacenter;
import com.materna.federatedcloudsim.FederatedHost;
import com.materna.federatedcloudsim.FederatedVm;
import com.materna.federatedcloudsim.core.FederatedCloudSimTags;
import com.materna.federatedcloudsim.core.FederatedConfigurationException;
import com.materna.federatedcloudsim.federation.ServiceProviderRegistry;
import com.materna.federatedcloudsim.federation.ServiceRequest;
import com.materna.federatedcloudsim.federation.sla.SlaCsp;
import com.materna.federatedcloudsim.scheduling.datacenter.FederatedVmAllocationPolicy;


/**
 * This class is a CloudBrokerStrategy that evaluates the RelianceScope metric.
 * This metric rates CloudServiceProvider by downtime and sla-breaches they
 * produce on federated Virtual Machines. The Strategy elects the
 * CloudServiceProvider that is the most reliable according to this metric.
 * 
 * The Strategy accepts migration requests, if the a suitable host is found by
 * evoking the DataCenterBrokers getDc(fedVm) and the related
 * VmAllocationPolicies findDestinationHost(fedVm) method.
 * 
 * @author Marcel Krger
 * 
 *         Parameter wich can be set via XML: -
 *         relianceDangerPercentageThreshold (float; default: 0.7f) -
 *         relianceDowntimeShareThreshold (float; default: 0.3f) -
 *         penaltyAvoidanceFactor (float; default: 0.7f) -
 *         downtimeAvoidanceFactor (float; default: 0.3f) -
 *         exponentialSmoothingCoefficient (float; default: 0.8f) -
 *         initialRelianceSope (float; default: 0.8f)
 *
 */
public class CloudBrokerStrategyRelianceScope extends CloudBrokerStrategy {

	private Map<FederatedVm, List<CloudServiceProvider>> vmsForMigration;
	private Map<CloudServiceProvider, Float> relianceScopeMap;
	private Map<CloudServiceProvider, TreeMap<FederatedVm, Long>> dtMemory;
	private Map<CloudServiceProvider, TreeMap<FederatedVm, Integer>> dtBrokenMemory;
	private Map<CloudServiceProvider, TreeMap<FederatedVm, Integer>> ramBrokenMemory;
	private Map<CloudServiceProvider, TreeMap<FederatedVm, Integer>> cpuBrokenMemory;
	private Map<CloudServiceProvider, Integer> numberPercentalRelianceValues;

	private Float relianceDangerPercentageThreshold = 0.7f;
	private Float relianceDowntimeShareThreshold = 0.3f;
	// the avoidance factors sum up to 1 and weight the 2 parts of reliance:
	// no penalties or no downtime
	private Float penaltyAvoidanceFactor = 0.7f;
	private Float downtimeAvoidanceFactor = 0.3f;
	private Float exponentialSmoothingCoefficient = 0.65f;
	private Float initialRelianceScope = 0.8f;

	public CloudBrokerStrategyRelianceScope() {
		super(false);
		vmsForMigration = new TreeMap<FederatedVm, List<CloudServiceProvider>>();
		relianceScopeMap = new TreeMap<CloudServiceProvider, Float>();
		numberPercentalRelianceValues = new TreeMap<CloudServiceProvider, Integer>();
	}
	
	public CloudBrokerStrategyRelianceScope(boolean auctionOn) {
		super(auctionOn);
		vmsForMigration = new TreeMap<FederatedVm, List<CloudServiceProvider>>();
		relianceScopeMap = new TreeMap<CloudServiceProvider, Float>();
		numberPercentalRelianceValues = new TreeMap<CloudServiceProvider, Integer>();
	}

	@Override
	public CloudServiceProvider getCsp(ServiceRequest sr, int evSource) {
		// first CSP with existing SLA is taken
		Set<SlaCsp> cspSlas = getCb().getCsp().getSlaManager().getUtilizedServiceSlas(sr.getSla().getService());
		Iterator<SlaCsp> iterator = cspSlas != null ? cspSlas.iterator() : null;
		CloudServiceProvider csp;
		while (iterator != null && iterator.hasNext()) {
			csp = iterator.next().getProviderCsp();
			// Event sent by own DC-Broker -> schedule to next available CSP
			if ((evSource == getCb().getCsp().getDcBroker().getId())) {
				return csp;
			}
			// Event sent by other CSPs CB -> execution rejected, schedule to
			// next CSP
			else if ((csp.getCb().getId() == evSource) && iterator.hasNext()) {
				csp = iterator.next().getProviderCsp();
				return csp;
			}
		}
		return null;
	}

	@Override
	public CloudServiceProvider findCsp(FederatedVm vm) {
		boolean foundCspToAsk = false;

		if (!vmsForMigration.containsKey(vm)) {
			vmsForMigration.put(vm, new ArrayList<CloudServiceProvider>());
			vmsForMigration.get(vm).add(this.getCb().getCsp());
		}

		List<CloudServiceProvider> sortedCspList = new ArrayList<>();
		sortedCspList.addAll(sortMapDescendingByScopeValue(getRelianceScopeMap()));

		for (CloudServiceProvider csp : sortedCspList) {
			if (!vmsForMigration.get(vm).contains(csp) && !foundCspToAsk) {
				foundCspToAsk = true;
				vmsForMigration.get(vm).add(csp);
				return csp;
			}
		}
		vmsForMigration.remove(vm);
		return null;
	}

	@Override
	public CloudServiceProvider handleRejectedMigrationRequest(FederatedVm vm) {
		boolean foundCspToAsk = false;
		if (vmsForMigration.containsKey(vm)) {

			for (CloudServiceProvider csp : sortMapDescendingByScopeValue(getRelianceScopeMap())) {
				if (!vmsForMigration.get(vm).contains(csp) && !foundCspToAsk) {
					foundCspToAsk = true;
					vmsForMigration.get(vm).add(csp);
					return csp;
				}
			}
		}
		vmsForMigration.remove(vm);
		return null;
	}

	@Override
	public boolean handleMigrationRequest(FederatedVm fedVm) {
		FederatedHost fedHost = null;

		FederatedDatacenter possDc = this.getCb().getCsp().getDcBroker().getStrategy().getDc(fedVm);

		if (possDc != null) {
			FederatedVmAllocationPolicy possAllocPolicy = (FederatedVmAllocationPolicy) possDc.getVmAllocationPolicy();
			fedHost = possAllocPolicy.findDestinationHost(fedVm);
		}

		return fedHost != null ? true : false;
	}

	@Override
	public void handleFedMigrationAnswer(FederatedVm fedVm, SimEvent ev) {
		vmsForMigration.remove(fedVm);
	}

	@Override
	public boolean acceptRemoteServiceRequest(ServiceRequest sr) {
		// TODO Auto-generated method stub
		return false;
	}

	@Override
	public double getRescheduleDelay(ServiceRequest sr) {
		// TODO Auto-generated method stub
		return 0;
	}

	public void updateRelianceScope() {
		for (CloudServiceProvider csp : getRelianceScopeMap().keySet()) {
			int migratedVmsQuantity = csp.getCb().getVmsMigrated().size();
			int properExecutedMigratedVms = 0;
			int unpunishedMigratedVms = 0;
			for (FederatedVm fedVm : csp.getCb().getVmsMigrated().keySet()) {
				boolean properExecuted = false;
				if (fedVm.getActualDownTime() == fedVm.getOldDowntime())
					properExecuted = true;
				if (dtMemory.get(csp).containsKey(fedVm)) {
					if (fedVm.getActualDownTime() <= dtMemory.get(csp).get(fedVm))
						properExecuted = true;
				} else {
					Float slaDangerPercentage = fedVm.getActualDownTime() / fedVm.getMaxDowntime();
					Float downtimeParticipationPercentage = getDowntimeShare(csp, fedVm);
					if (slaDangerPercentage <= relianceDangerPercentageThreshold
							|| downtimeParticipationPercentage <= relianceDowntimeShareThreshold) {
						properExecuted = true;
					}
				}
				if (properExecuted)
					properExecutedMigratedVms++;
				boolean penaltyOccured = false;
				if (dtBrokenMemory.get(csp).containsKey(fedVm)) {
					if (dtBrokenMemory.get(csp).get(fedVm) > fedVm.getDTBrokenCount()) {
						penaltyOccured = true;
					}
				}
				if (cpuBrokenMemory.get(csp).containsKey(fedVm)) {
					if (cpuBrokenMemory.get(csp).get(fedVm) > fedVm.getCpuBrokenCount()) {
						penaltyOccured = true;
					}
				}
				if (ramBrokenMemory.get(csp).containsKey(fedVm)) {
					if (ramBrokenMemory.get(csp).get(fedVm) > fedVm.getRamBrokenCount()) {
						penaltyOccured = true;
					}
				}
				if (!penaltyOccured)
					unpunishedMigratedVms++;
			}
			Float percentalReliance;
			if (migratedVmsQuantity != 0) {
				percentalReliance = (unpunishedMigratedVms / migratedVmsQuantity) * penaltyAvoidanceFactor
						+ (properExecutedMigratedVms / migratedVmsQuantity) * downtimeAvoidanceFactor;

				Float newWeightedAverage = exponentialSmoothingCoefficient * percentalReliance
						+ (1 - exponentialSmoothingCoefficient) * relianceScopeMap.get(csp);
				relianceScopeMap.put(csp, newWeightedAverage);
			}
		}
		dtMemory.clear();
		dtBrokenMemory.clear();
		cpuBrokenMemory.clear();
		ramBrokenMemory.clear();
		for (CloudServiceProvider csp : getRelianceScopeMap().keySet()) {
			TreeMap<FederatedVm, Long> actualDtMemory = new TreeMap<FederatedVm, Long>();
			TreeMap<FederatedVm, Integer> actualDtBrokenMemory = new TreeMap<FederatedVm, Integer>();
			TreeMap<FederatedVm, Integer> actualCpuBrokenMemory = new TreeMap<FederatedVm, Integer>();
			TreeMap<FederatedVm, Integer> actualRamBrokenMemory = new TreeMap<FederatedVm, Integer>();
			for (FederatedVm fedVm : csp.getCb().getVmsMigrated().keySet()) {
				actualDtMemory.put(fedVm, fedVm.getActualDownTime());
				actualDtBrokenMemory.put(fedVm, fedVm.getDTBrokenCount());
				actualCpuBrokenMemory.put(fedVm, fedVm.getCpuBrokenCount());
				actualRamBrokenMemory.put(fedVm, fedVm.getRamBrokenCount());
			}
			dtMemory.put(csp, actualDtMemory);
			dtBrokenMemory.put(csp, actualDtBrokenMemory);
			cpuBrokenMemory.put(csp, actualCpuBrokenMemory);
			ramBrokenMemory.put(csp, actualRamBrokenMemory);
		}
	}

	public List<CloudServiceProvider> sortMapDescendingByScopeValue(Map<CloudServiceProvider, Float> relianceScopeMap) {
		Map<CloudServiceProvider, Float> cacheMap = new TreeMap<CloudServiceProvider, Float>(relianceScopeMap);
		List<CloudServiceProvider> resultList = new ArrayList<>();
		CloudServiceProvider mostReliable = null;
		while (!cacheMap.isEmpty()) {
			Float highestScope = Float.MIN_VALUE;
			for (Map.Entry<CloudServiceProvider, Float> entry : cacheMap.entrySet()) {
				if (entry.getValue() == 0 && highestScope == Float.MIN_VALUE) {
					highestScope = entry.getValue();
					mostReliable = entry.getKey();
				}
				if (entry.getValue() > highestScope) {
					highestScope = entry.getValue();
					mostReliable = entry.getKey();
				}
			}
			cacheMap.remove(mostReliable);
			resultList.add(mostReliable);
			mostReliable = null;
		}
		return resultList;
	}

	public float getDowntimeShare(CloudServiceProvider csp, FederatedVm fedVm) {
		Long federationDowntime = 0l;
		Map<CloudServiceProvider, Long> downtimeMap = fedVm.getDowntimeMap();
		if (downtimeMap.size() == 0) {
			return (fedVm.getActualDownTime() - fedVm.getOldDowntime()) / fedVm.getActualDownTime();
		}
		federationDowntime = (fedVm.getActualDownTime() - fedVm.getOldDowntime());
		for (CloudServiceProvider cspInDtMap : downtimeMap.keySet()) {
			if (cspInDtMap == csp) {
				federationDowntime += downtimeMap.get(csp);
			}
		}
		return federationDowntime / fedVm.getActualDownTime();
	}

	/**
	 * @return the relianceScopeMap
	 */
	public Map<CloudServiceProvider, Float> getRelianceScopeMap() {
		return relianceScopeMap;
	}

	@Override
	public void initializeParameter() {
		for (CloudServiceProvider csp : ServiceProviderRegistry.getInstance().getCspList()) {
			if (csp != getCb().getCsp()) {
				relianceScopeMap.put(csp, initialRelianceScope);
				dtMemory = new TreeMap<CloudServiceProvider, TreeMap<FederatedVm, Long>>();
				dtMemory.put(csp, null);
				dtBrokenMemory = new TreeMap<CloudServiceProvider, TreeMap<FederatedVm, Integer>>();
				dtBrokenMemory.put(csp, null);
				ramBrokenMemory = new TreeMap<CloudServiceProvider, TreeMap<FederatedVm, Integer>>();
				ramBrokenMemory.put(csp, null);
				cpuBrokenMemory = new TreeMap<CloudServiceProvider, TreeMap<FederatedVm, Integer>>();
				cpuBrokenMemory.put(csp, null);
				numberPercentalRelianceValues.put(csp, 0);
			}
		}
	}

	/**
	 * Method is called after a instance of this class is created. It sets all
	 * configured parameter. otherwise a parameter will be set to his default
	 * value. \n
	 * 
	 * Parameter wich can be set via XML:\n - relianceDangerPercentageThreshold
	 * (float; default: 0.7f); \n - relianceDowntimeShareThreshold (float;
	 * default: 0.3f); \n - penaltyAvoidanceFactor (float; default: 0.7f); \n -
	 * downtimeAvoidanceFactor (float; default: 0.3f); \n -
	 * exponentialSmoothingCoefficient (float; default:0.8f); \ -
	 * initialRelianceSope (float; default: 0.8f); \n
	 */
	@Override
	public void configure(Map<String, String> params) throws FederatedConfigurationException {
		for (String key : params.keySet()) {
			switch (key) {
			case "relianceDangerPercentageThreshold":
				relianceDangerPercentageThreshold = Float.parseFloat(params.get("relianceDangerPercentageThreshold"));
				break;
			case "relianceDowntimeShareThreshold":
				relianceDowntimeShareThreshold = Float.parseFloat(params.get("relianceDowntimeShareThreshold"));
				break;
			case "penaltyAvoidanceFactor":
				penaltyAvoidanceFactor = Float.parseFloat(params.get("penaltyAvoidanceFactor"));
				break;
			case "downtimeAvoidanceFactor":
				downtimeAvoidanceFactor = Float.parseFloat(params.get("downtimeAvoidanceFactor"));
				break;
			case "exponentialSmoothingCoefficient":
				exponentialSmoothingCoefficient = Float.parseFloat(params.get("exponentialSmoothingCoefficient"));
				break;
			case "initialRelianceScope":
				initialRelianceScope = Float.parseFloat(params.get("initialRelianceScope"));
				break;
			default:
				throw new FederatedConfigurationException("Wrong parameter: " + key);
			}
		}
	}

}
